# -*- coding: utf-8 -*-
"""Stdout, stderr and argv support for unicode."""
#
# (C) Pywikibot team, 2012-2018
#
##############################################
# Support for unicode in windows cmd.exe
# Posted on Stack Overflow [1], available under CC-BY-SA 3.0 [2]
#
# Question: "Windows cmd encoding change causes Python crash" [3] by Alex [4],
# Answered [5] by David-Sarah Hopwood [6].
#
# [1] https://stackoverflow.com
# [2] https://creativecommons.org/licenses/by-sa/3.0/
# [3] https://stackoverflow.com/questions/878972
# [4] https://stackoverflow.com/users/85185
# [5] https://stackoverflow.com/a/3259271/118671
# [6] https://stackoverflow.com/users/393146
#
################################################
#
# Licensed under both CC-BY-SA and the MIT license.
#
################################################
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

import codecs
import sys

from ctypes import Structure, byref, c_int, create_unicode_buffer, sizeof
from ctypes import c_void_p as LPVOID
from io import IOBase, UnsupportedOperation

from pywikibot.tools import PY2, UnicodeType

OSWIN32 = (sys.platform == 'win32')

stdin = sys.stdin
stdout = sys.stdout
stderr = sys.stderr
argv = sys.argv

original_stderr = sys.stderr

if OSWIN32:
    from ctypes import WINFUNCTYPE, windll, POINTER, WinError
    from ctypes.wintypes import (BOOL, DWORD, HANDLE, LPCWSTR, LPWSTR,
                                 SHORT, ULONG, UINT, WCHAR)

try:
    ReadConsoleW = WINFUNCTYPE(BOOL, HANDLE, LPVOID, DWORD, POINTER(DWORD),
                               LPVOID)(('ReadConsoleW', windll.kernel32))
    WriteConsoleW = WINFUNCTYPE(BOOL, HANDLE, LPWSTR, DWORD, POINTER(DWORD),
                                LPVOID)(('WriteConsoleW', windll.kernel32))
except NameError:
    ReadConsoleW = WriteConsoleW = None


class UnicodeInput(IOBase):

    """Unicode terminal input class."""

    def __init__(self, hConsole, name, bufsize=1024):
        """Initialize the input stream."""
        self._hConsole = hConsole
        self.bufsize = bufsize
        self.buffer = create_unicode_buffer(bufsize)
        self.name = name
        self.encoding = 'utf-8'

    def readline(self):
        """Read one line from the input."""
        maxnum = DWORD(self.bufsize - 1)
        numrecv = DWORD(0)
        result = ReadConsoleW(self._hConsole, self.buffer, maxnum,
                              byref(numrecv), None)
        if not result:
            raise Exception('stdin failure')
        data = self.buffer.value[:numrecv.value]
        if PY2:
            return data.encode(self.encoding)
        else:
            return data


class UnicodeOutput(IOBase):

    """Unicode terminal output class."""

    def __init__(self, hConsole, stream, fileno, name):
        """Initialize the output stream."""
        self._hConsole = hConsole
        self._stream = stream
        self._fileno = fileno
        self.softspace = False
        self.mode = 'w'
        self.encoding = 'utf-8'
        self.name = name
        self.flush()

    def fileno(self):
        """Return the fileno."""
        return self._fileno

    def flush(self):
        """Flush the stream."""
        if self._hConsole is None:
            try:
                self._stream.flush()
            except Exception as e:
                _complain('%s.flush: %r from %r'
                          % (self.name, e, self._stream))
                raise

    def write(self, text):
        """Write the text to the output."""
        try:
            if self._hConsole is None:
                if PY2 and isinstance(text, UnicodeType):
                    text = text.encode('utf-8')
                self._stream.write(text)
            else:
                if not isinstance(text, UnicodeType):
                    text = bytes(text).decode('utf-8')
                remaining = len(text)
                while remaining > 0:
                    n = DWORD(0)
                    # There is a shorter-than-documented limitation on the
                    # length of the string passed to WriteConsoleW (see
                    # <https://tahoe-lafs.org/trac/tahoe-lafs/ticket/1232>.
                    retval = WriteConsoleW(self._hConsole, text,
                                           min(remaining, 10000),
                                           byref(n), None)
                    if retval == 0 or n.value == 0:
                        raise IOError('WriteConsoleW returned %r, n.value = %r'
                                      % (retval, n.value))
                    remaining -= n.value
                    if remaining == 0:
                        break
                    text = text[n.value:]
        except Exception as e:
            _complain('%s.write: %r' % (self.name, e))
            raise

    def writelines(self, lines):
        """Write a list of lines by using write."""
        try:
            for line in lines:
                self.write(line)
        except Exception as e:
            _complain('%s.writelines: %r' % (self.name, e))
            raise


def old_fileno(std_name):
    """Return the fileno or None if that doesn't work."""
    # some environments like IDLE don't support the fileno operation
    # handle those like std streams which don't have fileno at all
    std = getattr(sys, 'std{0}'.format(std_name))
    if hasattr(std, 'fileno'):
        try:
            return std.fileno()
        except UnsupportedOperation:
            pass


# If any exception occurs in this code, try to print it on stderr,
# which makes for frustrating debugging if stderr is directed to our wrapper.
# So be paranoid about catching errors and reporting them to original_stderr,
# so that we can at least see them.
def _complain(message):
    print(isinstance(message, str) and message or repr(message),
          file=original_stderr)


def register_cp65001():
    """Register codecs cp65001 as utf-8."""
    # Work around <http://bugs.python.org/issue6058>.
    codecs.register(lambda name: name == 'cp65001'
                    and codecs.lookup('utf-8') or None)


def force_truetype_console(h_stdout):
    """Force the console to use a TrueType font (Vista+)."""
    TMPF_TRUETYPE = 0x04
    LF_FACESIZE = 32

    class COORD(Structure):
        _fields_ = [('X', SHORT),
                    ('Y', SHORT)]

    class CONSOLE_FONT_INFOEX(Structure):
        _fields_ = [('cbSize', ULONG),
                    ('nFont', DWORD),
                    ('dwFontSize', COORD),
                    ('FontFamily', UINT),
                    ('FontWeight', UINT),
                    ('FaceName', WCHAR * LF_FACESIZE)]

    try:
        GetCurrentConsoleFontEx = WINFUNCTYPE(
            BOOL,
            HANDLE,  # hConsoleOutput
            BOOL,    # bMaximumWindow
            POINTER(CONSOLE_FONT_INFOEX),  # lpConsoleCurrentFontEx
        )(('GetCurrentConsoleFontEx', windll.kernel32))

        SetCurrentConsoleFontEx = WINFUNCTYPE(
            BOOL,
            HANDLE,  # hConsoleOutput
            BOOL,    # bMaximumWindow
            POINTER(CONSOLE_FONT_INFOEX),  # lpConsoleCurrentFontEx
        )(('SetCurrentConsoleFontEx', windll.kernel32))
    except AttributeError:
        # pre Windows Vista. Return without doing anything.
        return

    current_font = CONSOLE_FONT_INFOEX()
    current_font.cbSize = sizeof(CONSOLE_FONT_INFOEX)

    if not GetCurrentConsoleFontEx(h_stdout, True, byref(current_font)):
        WinError()

    truetype_font = (current_font.FontFamily & TMPF_TRUETYPE)

    if not truetype_font:
        new_font = CONSOLE_FONT_INFOEX()
        new_font.cbSize = sizeof(CONSOLE_FONT_INFOEX)
        new_font.FaceName = 'Lucida Console'

        if not SetCurrentConsoleFontEx(h_stdout, True, byref(new_font)):
            WinError()


def get_unicode_console():
    """
    Get Unicode console objects.

    @return: stdin, stdout, stderr, argv
    @rtype: tuple
    """
    # Make Unicode console output work independently of the current code page.
    # This also fixes http://bugs.python.org/issue1602.
    # Credit to Michael Kaplan
    # http://blogs.msdn.com/b/michkap/archive/2010/04/07/9989346.aspx
    # and TZOmegaTZIOY
    # https://stackoverflow.com/questions/878972/windows-cmd-encoding-change-causes-python-crash/1432462#1432462

    global stdin, stdout, stderr, argv

    if not OSWIN32:
        return stdin, stdout, stderr, argv

    try:
        # <https://msdn.microsoft.com/en-us/library/ms683231(VS.85).aspx>
        # HANDLE WINAPI GetStdHandle(DWORD nStdHandle);
        # returns INVALID_HANDLE_VALUE, NULL, or a valid handle
        #
        # <https://msdn.microsoft.com/en-us/library/aa364960(VS.85).aspx>
        # DWORD WINAPI GetFileType(DWORD hFile);
        #
        # <https://msdn.microsoft.com/en-us/library/ms683167(VS.85).aspx>
        # BOOL WINAPI GetConsoleMode(HANDLE hConsole, LPDWORD lpMode);

        GetStdHandle = WINFUNCTYPE(HANDLE, DWORD)(('GetStdHandle',
                                                   windll.kernel32))
        STD_INPUT_HANDLE = DWORD(-10)
        STD_OUTPUT_HANDLE = DWORD(-11)
        STD_ERROR_HANDLE = DWORD(-12)
        GetFileType = WINFUNCTYPE(DWORD, DWORD)(('GetFileType',
                                                 windll.kernel32))
        FILE_TYPE_CHAR = 0x0002
        FILE_TYPE_REMOTE = 0x8000
        GetConsoleMode = (WINFUNCTYPE(BOOL, HANDLE, POINTER(DWORD))
                          (('GetConsoleMode', windll.kernel32)))
        INVALID_HANDLE_VALUE = DWORD(-1).value

        def not_a_console(handle):
            """Return whether the handle is not to a console."""
            if handle == INVALID_HANDLE_VALUE or handle is None:
                return True
            return ((GetFileType(handle) & ~FILE_TYPE_REMOTE) != FILE_TYPE_CHAR
                    or GetConsoleMode(handle, byref(DWORD())) == 0)

        old_stdin_fileno = old_fileno('in')
        old_stdout_fileno = old_fileno('out')
        old_stderr_fileno = old_fileno('err')

        STDIN_FILENO = 0
        STDOUT_FILENO = 1
        STDERR_FILENO = 2
        real_stdin = (old_stdin_fileno == STDIN_FILENO)
        real_stdout = (old_stdout_fileno == STDOUT_FILENO)
        real_stderr = (old_stderr_fileno == STDERR_FILENO)

        if real_stdin:
            hStdin = GetStdHandle(STD_INPUT_HANDLE)
            if not_a_console(hStdin):
                real_stdin = False

        if real_stdout:
            hStdout = GetStdHandle(STD_OUTPUT_HANDLE)
            force_truetype_console(hStdout)
            if not_a_console(hStdout):
                real_stdout = False

        if real_stderr:
            hStderr = GetStdHandle(STD_ERROR_HANDLE)
            force_truetype_console(hStderr)
            if not_a_console(hStderr):
                real_stderr = False

        if real_stdout or real_stderr:
            if real_stdin:
                stdin = UnicodeInput(hStdin, name='<Unicode console stdin>')

            if real_stdout:
                stdout = UnicodeOutput(hStdout, sys.stdout, STDOUT_FILENO,
                                       '<Unicode console stdout>')
            else:
                stdout = UnicodeOutput(None, sys.stdout, old_stdout_fileno,
                                       '<Unicode redirected stdout>')

            if real_stderr:
                stderr = UnicodeOutput(hStderr, sys.stderr, STDERR_FILENO,
                                       '<Unicode console stderr>')
            else:
                stderr = UnicodeOutput(None, sys.stderr, old_stderr_fileno,
                                       '<Unicode redirected stderr>')
    except Exception as e:
        _complain('exception {!r} while fixing up sys.stdout and sys.stderr'
                  .format(e))

    if not PY2:
        # no need to work around issue2128 since it's a Python 2 only issue
        return stdin, stdout, stderr, argv

    # While we're at it, let's unmangle the command-line arguments:

    # This works around <http://bugs.python.org/issue2128>.
    GetCommandLineW = WINFUNCTYPE(LPWSTR)(('GetCommandLineW', windll.kernel32))
    CommandLineToArgvW = (WINFUNCTYPE(POINTER(LPWSTR), LPCWSTR, POINTER(c_int))
                          (('CommandLineToArgvW', windll.shell32)))

    argc = c_int(0)
    argv_unicode = CommandLineToArgvW(GetCommandLineW(), byref(argc))

    argv = [argv_unicode[i].encode('utf-8') for i in range(0, argc.value)]

    if not hasattr(sys, 'frozen'):
        # If this is an executable produced by py2exe or bbfreeze, then it will
        # have been invoked directly. Otherwise, unicode_argv[0] is the Python
        # interpreter, so skip that.
        argv = argv[1:]

        # Also skip option arguments to the Python interpreter.
        while len(argv) > 0:
            arg = argv[0]
            if not arg.startswith(b'-') or arg == '-':
                break
            argv = argv[1:]
            if arg == '-m':
                # sys.argv[0] should really be the absolute path of the module
                # source, but never mind
                break
            if arg == '-c':
                argv[0] = '-c'
                break

    if argv == []:
        argv = ['']

    return stdin, stdout, stderr, argv


if OSWIN32:
    register_cp65001()



# -*- coding: utf-8 -*-
"""Module to determine the pywikibot version (tag, revision and date)."""
#
# (C) Pywikibot team, 2007-2019
#
# Distributed under the terms of the MIT license.
#
from __future__ import absolute_import, division, unicode_literals

import codecs
import datetime
from importlib import import_module
import json
import os
import socket
import subprocess
import sys
import time
import xml.dom.minidom

from distutils import log
from distutils.sysconfig import get_python_lib
from io import BytesIO
from warnings import warn

try:
    from setuptools import svn_utils
except ImportError:
    try:
        from setuptools_svn import svn_utils
    except ImportError as e:
        svn_utils = e

try:
    import pathlib
except ImportError:
    try:
        import pathlib2 as pathlib
    except ImportError as e:
        pathlib = e

import pywikibot

from pywikibot import config2 as config
from pywikibot.tools import deprecated, PY2, UnicodeType

cache = None
_logger = 'version'


class ParseError(Exception):

    """Parsing went wrong."""


def _get_program_dir():
    _program_dir = os.path.normpath(
        os.path.split(os.path.dirname(__file__))[0])
    return _program_dir


def get_toolforge_hostname():
    """Get hostname of the current Toolforge host.

    @return: The hostname of the currently running host,
             if it is in Wikimedia Toolforge; otherwise return None.
    @rtype: str or None
    """
    if socket.getfqdn().endswith('.tools.eqiad.wmflabs'):
        return socket.gethostname()
    return None


def getversion(online=True):
    """Return a pywikibot version string.

    @param online: (optional) Include information obtained online
    """
    data = dict(getversiondict())  # copy dict to prevent changes in 'cache'
    data['cmp_ver'] = 'n/a'

    if online:
        try:
            hsh3 = getversion_onlinerepo('tags/stable')
            hsh2 = getversion_onlinerepo()
            hsh1 = data['hsh']
            data['cmp_ver'] = 'UNKNOWN' if not hsh1 else (
                'OUTDATED' if hsh1 not in (hsh2, hsh3) else 'ok')
        except Exception:
            pass

    data['hsh'] = data['hsh'][:7]  # make short hash from full hash
    return '%(tag)s (%(hsh)s, %(rev)s, %(date)s, %(cmp_ver)s)' % data


def getversiondict():
    """Get version info for the package.

    @return:
        - tag (name for the repository),
        - rev (current revision identifier),
        - date (date of current revision),
        - hash (git hash for the current revision)
    @rtype: C{dict} of four C{str}
    """
    global cache
    if cache:
        return cache

    _program_dir = _get_program_dir()
    exceptions = {}

    for vcs_func in (getversion_git,
                     getversion_svn_setuptools,
                     getversion_svn,
                     getversion_nightly,
                     getversion_package):
        try:
            (tag, rev, date, hsh) = vcs_func(_program_dir)
        except Exception as e:
            exceptions[vcs_func] = e
        else:
            break
    else:
        # nothing worked; version unknown (but suppress exceptions)
        # the value is most likely '$Id' + '$', it means that
        # pywikibot was imported without using version control at all.
        tag, rev, date, hsh = (
            '', '-1 (unknown)', '0 (unknown)', '(unknown)')

    # git and svn can silently fail, as it may be a nightly.
    if getversion_package in exceptions:
        warn('Unable to detect version; exceptions raised:\n%r'
             % exceptions, UserWarning)
    elif exceptions:
        pywikibot.debug('version algorithm exceptions:\n%r'
                        % exceptions, _logger)

    if isinstance(date, UnicodeType):
        datestring = date
    elif isinstance(date, time.struct_time):
        datestring = time.strftime('%Y/%m/%d, %H:%M:%S', date)
    else:
        warn('Unable to detect package date', UserWarning)
        datestring = '-2 (unknown)'

    cache = {'tag': tag, 'rev': rev, 'date': datestring, 'hsh': hsh}
    return cache


@deprecated('getversion_svn_setuptools', since='20150405')
def svn_rev_info(path):
    """Fetch information about the current revision of an Subversion checkout.

    @param path: directory of the Subversion checkout
    @return:
        - tag (name for the repository),
        - rev (current Subversion revision identifier),
        - date (date of current revision),
    @rtype: C{tuple} of two C{str} and a C{time.struct_time}
    """
    if not os.path.isdir(os.path.join(path, '.svn')):
        path = os.path.join(path, '..')

    _program_dir = path
    filename = os.path.join(_program_dir, '.svn/entries')
    if os.path.isfile(filename):
        with open(filename) as entries:
            version = entries.readline().strip()
            if version != '12':
                for i in range(3):
                    entries.readline()
                tag = entries.readline().strip()
                t = tag.split('://', 1)
                t[1] = t[1].replace('svn.wikimedia.org/svnroot/pywikipedia/',
                                    '')
                tag = '[{0}] {1}'.format(*t)
                for i in range(4):
                    entries.readline()
                date = time.strptime(entries.readline()[:19],
                                     '%Y-%m-%dT%H:%M:%S')
                rev = entries.readline()[:-1]
                return tag, rev, date

    # We haven't found the information in entries file.
    # Use sqlite table for new entries format
    from sqlite3 import dbapi2 as sqlite
    con = sqlite.connect(os.path.join(_program_dir, '.svn/wc.db'))
    cur = con.cursor()
    cur.execute("""select
local_relpath, repos_path, revision, changed_date, checksum from nodes
order by revision desc, changed_date desc""")
    name, tag, rev, date, checksum = cur.fetchone()
    cur.execute('select root from repository')
    tag, = cur.fetchone()
    con.close()
    tag = os.path.split(tag)[1]
    date = time.gmtime(date / 1000000)
    return tag, rev, date


def github_svn_rev2hash(tag, rev):
    """Convert a Subversion revision to a Git hash using Github.

    @param tag: name of the Subversion repo on Github
    @param rev: Subversion revision identifier
    @return: the git hash
    @rtype: str
    """
    from pywikibot.comms import http

    uri = 'https://github.com/wikimedia/%s/!svn/vcc/default' % tag
    request = http.fetch(uri=uri, method='PROPFIND',
                         body="<?xml version='1.0' encoding='utf-8'?>"
                              '<propfind xmlns=\"DAV:\"><allprop/></propfind>',
                         headers={'label': str(rev),
                                  'user-agent': 'SVN/1.7.5 {pwb}'})

    dom = xml.dom.minidom.parse(BytesIO(request.raw))
    hsh = dom.getElementsByTagName('C:git-commit')[0].firstChild.nodeValue
    date = dom.getElementsByTagName('S:date')[0].firstChild.nodeValue
    date = time.strptime(date[:19], '%Y-%m-%dT%H:%M:%S')
    return hsh, date


def getversion_svn_setuptools(path=None):
    """Get version info for a Subversion checkout using setuptools.

    @param path: directory of the Subversion checkout
    @return:
        - tag (name for the repository),
        - rev (current Subversion revision identifier),
        - date (date of current revision),
        - hash (git hash for the Subversion revision)
    @rtype: C{tuple} of three C{str} and a C{time.struct_time}
    """
    if isinstance(svn_utils, Exception):
        raise svn_utils
    tag = 'pywikibot-core'
    _program_dir = path or _get_program_dir()
    svninfo = svn_utils.SvnInfo(_program_dir)
    # suppress warning
    old_level = log.set_threshold(log.ERROR)
    rev = svninfo.get_revision()
    log.set_threshold(old_level)
    if not isinstance(rev, int):
        raise TypeError('SvnInfo.get_revision() returned type %s' % type(rev))
    if rev < 0:
        raise ValueError('SvnInfo.get_revision() returned %d' % rev)
    if rev == 0:
        raise ParseError('SvnInfo: invalid workarea')
    hsh, date = github_svn_rev2hash(tag, rev)
    rev = 's%s' % rev
    return (tag, rev, date, hsh)


@deprecated('getversion_svn_setuptools', since='20150405')
def getversion_svn(path=None):
    """Get version info for a Subversion checkout.

    @param path: directory of the Subversion checkout
    @return:
        - tag (name for the repository),
        - rev (current Subversion revision identifier),
        - date (date of current revision),
        - hash (git hash for the Subversion revision)
    @rtype: C{tuple} of three C{str} and a C{time.struct_time}
    """
    _program_dir = path or _get_program_dir()
    tag, rev, date = svn_rev_info(_program_dir)
    hsh, date2 = github_svn_rev2hash(tag, rev)
    if date.tm_isdst >= 0 and date2.tm_isdst >= 0:
        assert date == date2, 'Date of version is not consistent'
    # date.tm_isdst is -1 means unknown state
    # compare its contents except daylight saving time status
    else:
        for i in range(date.n_fields - 1):
            assert date[i] == date2[i], 'Date of version is not consistent'

    rev = 's%s' % rev
    if (not date or not tag or not rev) and not path:
        raise ParseError
    return (tag, rev, date, hsh)


def getversion_git(path=None):
    """Get version info for a Git clone.

    @param path: directory of the Git checkout
    @return:
        - tag (name for the repository),
        - rev (current revision identifier),
        - date (date of current revision),
        - hash (git hash for the current revision)
    @rtype: C{tuple} of three C{str} and a C{time.struct_time}
    """
    _program_dir = path or _get_program_dir()
    cmd = 'git'
    try:
        subprocess.Popen([cmd], stdout=subprocess.PIPE).communicate()
    except OSError:
        # some windows git versions provide git.cmd instead of git.exe
        cmd = 'git.cmd'

    with open(os.path.join(_program_dir, '.git/config'), 'r') as f:
        tag = f.read()
    # Try 'origin' and then 'gerrit' as remote name; bail if can't find either.
    remote_pos = tag.find('[remote "origin"]')
    if remote_pos == -1:
        remote_pos = tag.find('[remote "gerrit"]')
    if remote_pos == -1:
        tag = '?'
    else:
        s = tag.find('url = ', remote_pos)
        e = tag.find('\n', s)
        tag = tag[(s + 6):e]
        t = tag.strip().split('/')
        tag = '[%s] %s' % (t[0][:-1], '-'.join(t[3:]))
    dp = subprocess.Popen([cmd, '--no-pager',
                           'log', '-1',
                           '--pretty=format:"%ad|%an|%h|%H|%d"'
                           '--abbrev-commit',
                           '--date=iso'],
                          cwd=_program_dir,
                          stdout=subprocess.PIPE)
    info, stderr = dp.communicate()
    info = info.decode(config.console_encoding).split('|')
    date = info[0][:-6]
    date = time.strptime(date.strip('"'), '%Y-%m-%d %H:%M:%S')
    dp = subprocess.Popen([cmd, 'rev-list', 'HEAD'],
                          cwd=_program_dir,
                          stdout=subprocess.PIPE)
    rev, stderr = dp.communicate()
    rev = 'g%s' % len(rev.splitlines())
    hsh = info[3]  # also stored in '.git/refs/heads/master'
    if (not date or not tag or not rev) and not path:
        raise ParseError
    return (tag, rev, date, hsh)


def getversion_nightly(path=None):
    """Get version info for a nightly release.

    @param path: directory of the uncompressed nightly.
    @return:
        - tag (name for the repository),
        - rev (current revision identifier),
        - date (date of current revision),
        - hash (git hash for the current revision)
    @rtype: C{tuple} of three C{str} and a C{time.struct_time}
    """
    if not path:
        path = _get_program_dir()

    with open(os.path.join(path, 'version')) as data:
        (tag, rev, date, hsh) = data.readlines()

    date = time.strptime(date[:19], '%Y-%m-%dT%H:%M:%S')

    if not date or not tag or not rev:
        raise ParseError
    return (tag, rev, date, hsh)


def getversion_package(path=None):
    """Get version info for an installed package.

    @param path: Unused argument
    @return:
        - tag: 'pywikibot/__init__.py'
        - rev: '-1 (unknown)'
        - date (date the package was installed locally),
        - hash (git hash for the current revision of 'pywikibot/__init__.py')
    @rtype: C{tuple} of four C{str}
    """
    hsh = ''
    date = get_module_mtime(pywikibot).timetuple()

    tag = 'pywikibot/__init__.py'
    rev = '-1 (unknown)'

    return (tag, rev, date, hsh)


def getversion_onlinerepo(path='branches/master'):
    """Retrieve current framework git hash from Gerrit."""
    from pywikibot.comms import http
    # Gerrit API responses include )]}' at the beginning,
    # make sure to strip it out
    buf = http.fetch(
        uri='https://gerrit.wikimedia.org/r/projects/pywikibot%2Fcore/' + path,
        headers={'user-agent': '{pwb}'}).text[4:]
    try:
        hsh = json.loads(buf)['revision']
        return hsh
    except Exception as e:
        raise ParseError(repr(e) + ' while parsing ' + repr(buf))


@deprecated('get_module_version, get_module_filename and get_module_mtime',
            since='20150221')
def getfileversion(filename):
    """Retrieve revision number of file.

    Extracts __version__ variable containing Id tag, without importing it.
    (thus can be done for any file)

    The version variable containing the Id tag is read and
    returned. Because it doesn't import it, the version can
    be retrieved from any file.
    @param filename: Name of the file to get version
    @type filename: str
    """
    _program_dir = _get_program_dir()
    __version__ = None
    mtime = None
    fn = os.path.join(_program_dir, filename)
    if os.path.exists(fn):
        with codecs.open(fn, 'r', 'utf-8') as f:
            for line in f.readlines():
                if line.find('__version__') == 0:
                    try:
                        exec(line)
                    except Exception:
                        pass
                    break
        stat = os.stat(fn)
        mtime = datetime.datetime.fromtimestamp(stat.st_mtime).isoformat(' ')
    if mtime and __version__:
        return '%s %s %s' % (filename, __version__[5:-1][:7], mtime)
    else:
        return None


def get_module_version(module):
    """
    Retrieve __version__ variable from an imported module.

    @param module: The module instance.
    @type module: module
    @return: The version hash without the surrounding text. If not present
        return None.
    @rtype: str or None
    """
    if hasattr(module, '__version__'):
        return module.__version__[5:-1]


def get_module_filename(module):
    """
    Retrieve filename from an imported pywikibot module.

    It uses the __file__ attribute of the module. If it's file extension ends
    with py and another character the last character is discarded when the py
    file exist.

    @param module: The module instance.
    @type module: module
    @return: The filename if it's a pywikibot module otherwise None.
    @rtype: str or None
    """
    if hasattr(module, '__file__') and os.path.exists(module.__file__):
        filename = module.__file__
        if PY2:
            filename = os.path.abspath(filename)
        if filename[-4:-1] == '.py' and os.path.exists(filename[:-1]):
            filename = filename[:-1]
        program_dir = _get_program_dir()
        if filename[:len(program_dir)] == program_dir:
            return filename


def get_module_mtime(module):
    """
    Retrieve the modification time from an imported module.

    @param module: The module instance.
    @type module: module
    @return: The modification time if it's a pywikibot module otherwise None.
    @rtype: datetime or None
    """
    filename = get_module_filename(module)
    if filename:
        return datetime.datetime.fromtimestamp(os.stat(filename).st_mtime)


def package_versions(modules=None, builtins=False, standard_lib=None):
    """Retrieve package version information.

    When builtins or standard_lib are None, they will be included only
    if a version was found in the package.

    @param modules: Modules to inspect
    @type modules: list of strings
    @param builtins: Include builtins
    @type builtins: Boolean, or None for automatic selection
    @param standard_lib: Include standard library packages
    @type standard_lib: Boolean, or None for automatic selection
    """
    if not modules:
        modules = sys.modules.keys()

    std_lib_dir = get_python_lib(standard_lib=True)

    root_packages = {key.split('.')[0] for key in modules}

    builtin_packages = {name.split('.')[0] for name in root_packages
                        if name in sys.builtin_module_names
                        or '_' + name in sys.builtin_module_names}

    # Improve performance by removing builtins from the list if possible.
    if builtins is False:
        root_packages = list(root_packages - builtin_packages)

    std_lib_packages = []

    paths = {}
    data = {}

    for name in root_packages:
        try:
            package = import_module(name)
        except ImportError as e:
            data[name] = {'name': name, 'err': e}
            continue

        info = {'package': package, 'name': name}

        if name in builtin_packages:
            info['type'] = 'builtins'

        if '__file__' in package.__dict__:
            # Determine if this file part is of the standard library.
            if os.path.normcase(package.__file__).startswith(
                    os.path.normcase(std_lib_dir)):
                std_lib_packages.append(name)
                if standard_lib is False:
                    continue
                info['type'] = 'standard libary'

            # Strip '__init__.py' from the filename.
            path = package.__file__
            if '__init__.py' in path:
                path = path[0:path.index('__init__.py')]

            if PY2:
                path = path.decode(sys.getfilesystemencoding())

            info['path'] = path
            assert path not in paths, 'Path of the package is in defined paths'
            paths[path] = name

        if '__version__' in package.__dict__:
            info['ver'] = package.__version__
        elif name.startswith('unicodedata'):
            info['ver'] = package.unidata_version

        # If builtins or standard_lib is None,
        # only include package if a version was found.
        if (builtins is None and name in builtin_packages) or \
                (standard_lib is None and name in std_lib_packages):
            if 'ver' in info:
                data[name] = info
            else:
                # Remove the entry from paths, so it isn't processed below
                del paths[info['path']]
        else:
            data[name] = info

    # Remove any pywikibot sub-modules which were loaded as a package.
    # e.g. 'wikipedia_family.py' is loaded as 'wikipedia'
    _program_dir = _get_program_dir()
    if isinstance(pathlib, Exception):
        dir_parts = _program_dir.split(os.sep)
    else:
        dir_parts = pathlib.Path(_program_dir).parts
    length = len(dir_parts)
    for path, name in paths.items():
        if isinstance(pathlib, Exception):
            lib_parts = os.path.normpath(path).split(os.sep)
        else:
            lib_parts = pathlib.Path(path).parts
        if dir_parts != lib_parts[:length]:
            continue
        if lib_parts[length] != '.tox':
            del data[name]

    return data



# -*- coding: utf-8 -*-
"""Title translate module."""
#
# (C) Pywikibot team, 2003-2018
#
# Distributed under the terms of the MIT license.
#
from __future__ import absolute_import, division, unicode_literals

import pywikibot
from pywikibot import date

from pywikibot import config
from pywikibot.tools import deprecated_args


@deprecated_args(family=None)
def translate(page=None, hints=(), auto=True, removebrackets=False,
              site=None):
    """
    Return a list of links to pages on other sites based on hints.

    Entries for single page titles list those pages. Page titles for entries
    such as "all:" or "xyz:" or "20:" are first built from the page title of
    'page' and then listed. When 'removebrackets' is True, a trailing pair of
    brackets and the text between them is removed from the page title.
    If 'auto' is true, known year and date page titles are autotranslated
    to all known target languages and inserted into the list.
    """
    result = set()

    assert page or site

    if site is None:
        site = page.site

    for h in hints:
        # argument may be given as -hint:xy where xy is a language code
        codes, _, newname = h.partition(':')
        if not newname:
            # if given as -hint:xy or -hint:xy:, assume that there should
            # be a page in language xy with the same title as the page
            # we're currently working on ...
            if page is None:
                continue
            newname = page.title(with_ns=False,
                                 without_brackets=removebrackets)
        if codes.isdigit():
            codes = site.family.languages_by_size[:int(codes)]
        elif codes == 'all':
            codes = site.family.languages_by_size
        elif codes in site.family.language_groups:
            codes = site.family.language_groups[codes]
        else:
            codes = codes.split(',')

        for newcode in codes:
            if newcode in site.languages():
                if newcode != site.code:
                    ns = page.namespace() if page else 0
                    x = pywikibot.Link(newname,
                                       site.getSite(code=newcode),
                                       default_namespace=ns)
                    result.add(x)
            elif config.verbose_output:
                pywikibot.output('Ignoring unknown language code {}'
                                 .format(newcode))

    # Autotranslate dates into all other languages, the rest will come from
    # existing interwiki links.
    if auto and page:
        # search inside all dictionaries for this link
        sitelang = page.site.lang
        dict_name, value = date.getAutoFormat(sitelang, page.title())
        if dict_name:
            pywikibot.output(
                'TitleTranslate: %s was recognized as %s with value %d'
                % (page.title(), dict_name, value))
            for entry_lang, entry in date.formats[dict_name].items():
                if entry_lang not in site.languages():
                    continue
                if entry_lang != sitelang:
                    newname = entry(value)
                    x = pywikibot.Link(
                        newname,
                        pywikibot.Site(code=entry_lang,
                                       fam=site.family))
                    result.add(x)
    return list(result)



# -*- coding: utf-8 -*-
"""IP address tools module."""
#
# (C) Pywikibot team, 2014-2020
#
# Distributed under the terms of the MIT license.
#
from __future__ import absolute_import, division, unicode_literals

from distutils.version import StrictVersion

from pywikibot.tools import ModuleDeprecationWrapper

_ipaddress_e = _ipaddr_e = _ipaddr_version = None

try:
    from ipaddress import ip_address
except ImportError as e:
    _ipaddress_e = e
    ip_address = None

if not ip_address:
    try:
        from ipaddr import __version__ as _ipaddr_version
    except ImportError as e:
        _ipaddr_e = e
    else:
        _ipaddr_version = StrictVersion(_ipaddr_version)
        if _ipaddr_version >= StrictVersion('2.1.10'):
            from ipaddr import IPAddress as ip_address  # noqa: N813
        else:
            _ipaddr_e = ImportError('ipaddr %s is broken.' % _ipaddr_version)

if not ip_address:
    raise ImportError('Importing ipaddr.IPAddress failed: {}\n'
                      'Importing ipaddress.ip_address failed: {}\n'
                      'Please install ipaddress.'
                      .format(_ipaddr_e, _ipaddress_e))


def is_IP(IP):
    """
    Verify the IP address provided is valid.

    No logging is performed. Use ip_address instead to catch errors.

    @param IP: IP address
    @type IP: str
    @rtype: bool
    """
    try:
        ip_address(IP)
        return True
    except ValueError:
        return False


wrapper = ModuleDeprecationWrapper(__name__)
wrapper._add_deprecated_attr('is_IP',
                             replacement_name='tools.is_IP',
                             future_warning=True,
                             since='20200120')



# -*- coding: utf-8 -*-
"""Objects representing Mediawiki log entries."""
#
# (C) Pywikibot team, 2007-2019
#
# Distributed under the terms of the MIT license.
#
from __future__ import absolute_import, division, unicode_literals

import pywikibot
from pywikibot.exceptions import Error, HiddenKeyError
from pywikibot.tools import deprecated, classproperty, UnicodeType

_logger = 'wiki'


class LogDict(dict):

    """
    Simple custom dict that raises custom Errors when a key is missing.

    HiddenKeyError is raised when the user does not have permission.
    KeyError is raised otherwise.

    It also logs debugging information when a key is missing.
    """

    def __missing__(self, key):
        """Debug when the key is missing."""
        pywikibot.debug('API log entry received:\n' + repr(self),
                        _logger)
        if ((key in ('ns', 'title', 'pageid', 'logpage', 'params', 'action')
             and 'actionhidden' in self)
                or (key == 'comment' and 'commenthidden' in self)
                or (key == 'user' and 'userhidden' in self)):
            raise HiddenKeyError(
                "Log entry ({0}) has a hidden '{1}' key and you don't have "
                'permission to view it.'.format(self._type, key))
        raise KeyError("Log entry (%s) has no '%s' key" % (self._type, key))


class LogEntry(object):

    """Generic log entry."""

    # Log type expected. None for every type, or one of the (letype) str :
    # block/patrol/etc...
    # Overridden in subclasses.
    _expected_type = None

    def __init__(self, apidata, site):
        """Initialize object from a logevent dict returned by MW API."""
        self.data = LogDict(apidata)
        self.site = site
        expected_type = self._expected_type
        if expected_type is not None and expected_type != self.type():
            raise Error('Wrong log type! Expecting %s, received %s instead.'
                        % (expected_type, self.type()))
        self.data._type = self.type()

    def __repr__(self):
        """Return a string representation of LogEntry object."""
        return '<{0}({1}, logid={2})>'.format(type(self).__name__,
                                              self.site.sitename, self.logid())

    def __hash__(self):
        """Combine site and logid as the hash."""
        return self.logid() ^ hash(self.site)

    def __eq__(self, other):
        """Compare if self is equal to other."""
        if not isinstance(other, LogEntry):
            pywikibot.debug("'{0}' cannot be compared with '{1}'"
                            .format(type(self).__name__, type(other).__name__),
                            _logger)
            return False
        return self.logid() == other.logid() and self.site == other.site

    def __ne__(self, other):
        """Compare if self is not equal to other."""
        return not self == other

    @property
    def _params(self):
        """
        Additional data for some log entry types.

        @rtype: dict or None
        """
        if 'params' in self.data:
            return self.data['params']
        else:  # try old mw style preceding mw 1.19
            return self.data[self._expected_type]

    def logid(self):
        """Return the id of the log entry."""
        return self.data['logid']

    def pageid(self):
        """Return the log id of the page handled by this log entry."""
        return self.data['pageid']

    def ns(self):
        """Return the namespace id of the page handled by this log entry."""
        return self.data['ns']

    @deprecated('page()', since='20150617')
    def title(self):
        """
        DEPRECATED: Alias for page().

        This is going to be replaced by just returning the title as a string
        instead of a Page instance.
        """
        return self.page()

    def page(self):
        """
        Page on which action was performed.

        @return: page on action was performed
        @rtype: pywikibot.Page
        """
        if not hasattr(self, '_page'):
            self._page = pywikibot.Page(self.site, self.data['title'])
        return self._page

    def type(self):
        """The type of this logentry."""
        return self.data['type']

    def action(self):
        """The action of this log entry."""
        return self.data['action']

    def user(self):
        """Return the user name doing this action."""
        # TODO use specific User class ?
        return self.data['user']

    def timestamp(self):
        """Timestamp object corresponding to event timestamp."""
        if not hasattr(self, '_timestamp'):
            self._timestamp = pywikibot.Timestamp.fromISOformat(
                self.data['timestamp'])
        return self._timestamp

    def comment(self):
        """Return the logentry's comment.

        @rtype: str
        """
        return self.data['comment']


class OtherLogEntry(LogEntry):

    """A log entry class for unspecified log events."""

    pass


class UserTargetLogEntry(LogEntry):

    """A log entry whose target is a user page."""

    def page(self):
        """Return the target user.

        This returns a User object instead of the Page object returned by the
        superclass method.

        @return: target user
        @rtype: pywikibot.User
        """
        if not hasattr(self, '_page'):
            self._page = pywikibot.User(self.site, self.data['title'])
        return self._page


class BlockEntry(LogEntry):

    """
    Block or unblock log entry.

    It might contain a block or unblock depending on the action. The duration,
    expiry and flags are not available on unblock log entries.
    """

    _expected_type = 'block'

    def __init__(self, apidata, site):
        """Initializer."""
        super(BlockEntry, self).__init__(apidata, site)
        # When an autoblock is removed, the "title" field is not a page title
        # See bug T19781
        pos = self.data['title'].find('#')
        self.isAutoblockRemoval = pos > 0
        if self.isAutoblockRemoval:
            self._blockid = int(self.data['title'][pos + 1:])

    def page(self):
        """
        Return the blocked account or IP.

        @return: the Page object of username or IP if this block action
            targets a username or IP, or the blockid if this log reflects
            the removal of an autoblock
        @rtype: pywikibot.Page or int
        """
        # TODO what for IP ranges ?
        if self.isAutoblockRemoval:
            return self._blockid
        else:
            return super(BlockEntry, self).page()

    def flags(self):
        """
        Return a list of (str) flags associated with the block entry.

        It raises an Error if the entry is an unblocking log entry.

        @return: list of flags strings
        @rtype: list
        """
        if self.action() == 'unblock':
            return []
        if not hasattr(self, '_flags'):
            self._flags = self._params['flags']
            # pre mw 1.19 returned a delimited string.
            if isinstance(self._flags, UnicodeType):
                if self._flags:
                    self._flags = self._flags.split(',')
                else:
                    self._flags = []
        return self._flags

    def duration(self):
        """
        Return a datetime.timedelta representing the block duration.

        @return: datetime.timedelta, or None if block is indefinite.
        """
        if not hasattr(self, '_duration'):
            if self.expiry() is None:
                self._duration = None
            else:
                # Doing the difference is easier than parsing the string
                self._duration = self.expiry() - self.timestamp()
        return self._duration

    def expiry(self):
        """
        Return a Timestamp representing the block expiry date.

        @rtype: pywikibot.Timestamp or None
        """
        if not hasattr(self, '_expiry'):
            details = self._params.get('expiry')
            if details:
                self._expiry = pywikibot.Timestamp.fromISOformat(details)
            else:
                self._expiry = None  # for infinite blocks
        return self._expiry


class RightsEntry(LogEntry):

    """Rights log entry."""

    _expected_type = 'rights'

    @property
    def oldgroups(self):
        """Return old rights groups."""
        params = self._params
        if 'old' in params:  # old mw style
            return params['old'].split(',') if params['old'] else []
        return params['oldgroups']

    @property
    def newgroups(self):
        """Return new rights groups."""
        params = self._params
        if 'new' in params:  # old mw style
            return params['new'].split(',') if params['new'] else []
        return params['newgroups']


class UploadEntry(LogEntry):

    """Upload log entry."""

    _expected_type = 'upload'

    def page(self):
        """
        Return FilePage on which action was performed.

        @rtype: pywikibot.FilePage
        """
        if not hasattr(self, '_page'):
            self._page = pywikibot.FilePage(self.site, self.data['title'])
        return self._page


class MoveEntry(LogEntry):

    """Move log entry."""

    _expected_type = 'move'

    @deprecated('target_ns.id', since='20150518')
    def new_ns(self):
        """Return namespace id of target page."""
        return self.target_ns.id

    @property
    def target_ns(self):
        """Return namespace object of target page."""
        # key has been changed in mw 1.19
        return self.site.namespaces[self._params['target_ns']
                                    if 'target_ns' in self._params
                                    else self._params['new_ns']]

    @deprecated('target_page', since='20150518')
    def new_title(self):
        """Return page object of the new title."""
        return self.target_page

    @property
    def target_title(self):
        """Return the target title."""
        # key has been changed in mw 1.19
        return (self._params['target_title']
                if 'target_title' in self._params
                else self._params['new_title'])

    @property
    def target_page(self):
        """
        Return target page object.

        @rtype: pywikibot.Page
        """
        if not hasattr(self, '_target_page'):
            self._target_page = pywikibot.Page(self.site, self.target_title)
        return self._target_page

    def suppressedredirect(self):
        """
        Return True if no redirect was created during the move.

        @rtype: bool
        """
        # Introduced in MW r47901
        return 'suppressedredirect' in self._params


class PatrolEntry(LogEntry):

    """Patrol log entry."""

    _expected_type = 'patrol'

    @property
    def current_id(self):
        """
        Return the current id.

        @rtype: int
        """
        # key has been changed in mw 1.19; try the new mw style first
        # sometimes it returns strs sometimes ints
        return int(self._params['curid']
                   if 'curid' in self._params else self._params['cur'])

    @property
    def previous_id(self):
        """
        Return the previous id.

        @rtype: int
        """
        # key has been changed in mw 1.19; try the new mw style first
        # sometimes it returns strs sometimes ints
        return int(self._params['previd']
                   if 'previd' in self._params else self._params['prev'])

    @property
    def auto(self):
        """Return auto patrolled."""
        return 'auto' in self._params and self._params['auto'] != 0


class LogEntryFactory(object):

    """
    LogEntry Factory.

    Only available method is create()
    """

    _logtypes = {
        'block': BlockEntry,
        'rights': RightsEntry,
        'upload': UploadEntry,
        'move': MoveEntry,
        'patrol': PatrolEntry,
    }

    def __init__(self, site, logtype=None):
        """
        Initializer.

        @param site: The site on which the log entries are created.
        @type site: BaseSite
        @param logtype: The log type of the log entries, if known in advance.
                        If None, the Factory will fetch the log entry from
                        the data to create each object.
        @type logtype: (letype) str : move/block/patrol/etc...
        """
        self._site = site
        if logtype is None:
            self._creator = self._createFromData
        else:
            # Bind a Class object to self._creator:
            # When called, it will initialize a new object of that class
            logclass = self.get_valid_entry_class(logtype)
            self._creator = lambda data: logclass(data, self._site)

    @classproperty
    @deprecated('Site.logtypes or LogEntryFactory.get_entry_class(logtype)',
                since='20160918')
    def logtypes(cls):
        """DEPRECATED LogEntryFactory class attribute of log types."""
        return cls._logtypes

    def create(self, logdata):
        """
        Instantiate the LogEntry object representing logdata.

        @param logdata: <item> returned by the api
        @type logdata: dict

        @return: LogEntry object representing logdata
        """
        return self._creator(logdata)

    def get_valid_entry_class(self, logtype):
        """
        Return the class corresponding to the @logtype string parameter.

        @return: specified subclass of LogEntry
        @rtype: LogEntry
        @raise KeyError: logtype is not valid
        """
        if logtype not in self._site.logtypes:
            raise KeyError('{} is not a valid logtype'.format(logtype))
        return LogEntryFactory.get_entry_class(logtype)

    @classmethod
    def get_entry_class(cls, logtype):
        """
        Return the class corresponding to the @logtype string parameter.

        @return: specified subclass of LogEntry
        @rtype: LogEntry
        @note: this class method cannot verify whether the given logtype
            already exits for a given site; to verify use Site.logtypes
            or use the get_valid_entry_class instance method instead.
        """
        if logtype not in cls._logtypes:
            if logtype in ('newusers', 'thanks'):
                bases = (UserTargetLogEntry, OtherLogEntry)
            else:
                bases = (OtherLogEntry, )
            classname = str(logtype.capitalize() + 'Entry'
                            if logtype is not None
                            else OtherLogEntry.__name__)
            cls._logtypes[logtype] = type(
                classname, bases, {'_expected_type': logtype})
        return cls._logtypes[logtype]

    def _createFromData(self, logdata):
        """
        Check for logtype from data, and creates the correct LogEntry.

        @param logdata: log entry data
        @type logdata: dict
        @rtype: LogEntry
        """
        try:
            logtype = logdata['type']
        except KeyError:
            pywikibot.debug('API log entry received:\n{0}'.format(logdata),
                            _logger)
            raise Error("Log entry has no 'type' key")
        return LogEntryFactory.get_entry_class(logtype)(logdata, self._site)


# For backward compatibility
ProtectEntry = LogEntryFactory.get_entry_class('protect')
DeleteEntry = LogEntryFactory.get_entry_class('delete')
ImportEntry = LogEntryFactory.get_entry_class('import')
NewUsersEntry = LogEntryFactory.get_entry_class('newusers')
ThanksEntry = LogEntryFactory.get_entry_class('thanks')



# -*- coding: utf-8 -*-
"""Objects representing Flow entities, like boards, topics, and posts."""
#
# (C) Pywikibot team, 2015-2020
#
# Distributed under the terms of the MIT license.
#
from __future__ import absolute_import, division, unicode_literals

import logging

from pywikibot.exceptions import NoPage, UnknownExtension, LockedPage
from pywikibot.page import BasePage, User
from pywikibot.tools import PY2, UnicodeType

if not PY2:
    from urllib.parse import urlparse, parse_qs
else:
    from urlparse import urlparse, parse_qs


logger = logging.getLogger('pywiki.wiki.flow')


# Flow page-like objects (boards and topics)
class FlowPage(BasePage):

    """
    The base page for the Flow extension.

    There should be no need to instantiate this directly.

    Subclasses must provide a _load() method to load and cache
    the object's internal data from the API.
    """

    def __init__(self, source, title=''):
        """Initializer.

        @param source: A Flow-enabled site or a Link or Page on such a site
        @type source: Site, pywikibot.page.Link, or pywikibot.page.Page
        @param title: normalized title of the page
        @type title: str

        @raises TypeError: incorrect use of parameters
        @raises ValueError: use of non-Flow-enabled Site
        """
        super(FlowPage, self).__init__(source, title)

        if not self.site.has_extension('Flow'):
            raise UnknownExtension('site is not Flow-enabled')

    def _load_uuid(self):
        """Load and save the UUID of the page."""
        self._uuid = self._load()['workflowId']

    @property
    def uuid(self):
        """Return the UUID of the page.

        @return: UUID of the page
        @rtype: str
        """
        if not hasattr(self, '_uuid'):
            self._load_uuid()
        return self._uuid

    def get(self, force=False, get_redirect=False):
        """Get the page's content."""
        if get_redirect or force:
            raise NotImplementedError

        # TODO: Return more useful data
        return self._data


class Board(FlowPage):

    """A Flow discussion board."""

    def _load(self, force=False):
        """Load and cache the Board's data, derived from its topic list."""
        if not hasattr(self, '_data') or force:
            self._data = self.site.load_board(self)
        return self._data

    def _parse_url(self, links):
        """Parse a URL retrieved from the API."""
        rule = links['fwd']
        parsed_url = urlparse(rule['url'])
        params = parse_qs(parsed_url.query)
        new_params = {}
        for key, value in params.items():
            if key != 'title':
                key = key.replace('topiclist_', '').replace('-', '_')
                if key == 'offset_dir':
                    new_params['reverse'] = (value == 'rev')
                else:
                    new_params[key] = value
        return new_params

    def topics(self, format='wikitext', limit=100, sort_by='newest',
               offset=None, offset_uuid='', reverse=False,
               include_offset=False, toc_only=False):
        """Load this board's topics.

        @param format: The content format to request the data in.
        @type format: str (either 'wikitext', 'html', or 'fixed-html')
        @param limit: The number of topics to fetch in each request.
        @type limit: int
        @param sort_by: Algorithm to sort topics by.
        @type sort_by: str (either 'newest' or 'updated')
        @param offset: The timestamp to start at (when sortby is 'updated').
        @type offset: Timestamp or equivalent str
        @param offset_uuid: The UUID to start at (when sortby is 'newest').
        @type offset_uuid: str (in the form of a UUID)
        @param reverse: Whether to reverse the topic ordering.
        @type reverse: bool
        @param include_offset: Whether to include the offset topic.
        @type include_offset: bool
        @param toc_only: Whether to only include information for the TOC.
        @type toc_only: bool
        @return: A generator of this board's topics.
        @rtype: generator of Topic objects
        """
        data = self.site.load_topiclist(self, format=format, limit=limit,
                                        sortby=sort_by, toconly=toc_only,
                                        offset=offset, offset_id=offset_uuid,
                                        reverse=reverse,
                                        include_offset=include_offset)
        while data['roots']:
            for root in data['roots']:
                topic = Topic.from_topiclist_data(self, root, data)
                yield topic
            cont_args = self._parse_url(data['links']['pagination'])
            data = self.site.load_topiclist(self, **cont_args)

    def new_topic(self, title, content, format='wikitext'):
        """Create and return a Topic object for a new topic on this Board.

        @param title: The title of the new topic (must be in plaintext)
        @type title: str
        @param content: The content of the topic's initial post
        @type content: str
        @param format: The content format of the value supplied for content
        @type format: str (either 'wikitext' or 'html')
        @return: The new topic
        @rtype: Topic
        """
        return Topic.create_topic(self, title, content, format)


class Topic(FlowPage):

    """A Flow discussion topic."""

    def _load(self, format='wikitext', force=False):
        """Load and cache the Topic's data."""
        if not hasattr(self, '_data') or force:
            self._data = self.site.load_topic(self, format)
        return self._data

    def _reload(self):
        """Forcibly reload the topic's root post."""
        self.root._load(load_from_topic=True)

    @classmethod
    def create_topic(cls, board, title, content, format='wikitext'):
        """Create and return a Topic object for a new topic on a Board.

        @param board: The topic's parent board
        @type board: Board
        @param title: The title of the new topic (must be in plaintext)
        @type title: str
        @param content: The content of the topic's initial post
        @type content: str
        @param format: The content format of the value supplied for content
        @type format: str (either 'wikitext' or 'html')
        @return: The new topic
        @rtype: Topic
        """
        data = board.site.create_new_topic(board, title, content, format)
        return cls(board.site, data['topic-page'])

    @classmethod
    def from_topiclist_data(cls, board, root_uuid, topiclist_data):
        """Create a Topic object from API data.

        @param board: The topic's parent Flow board
        @type board: Board
        @param root_uuid: The UUID of the topic and its root post
        @type root_uuid: str
        @param topiclist_data: The data returned by view-topiclist
        @type topiclist_data: dict
        @return: A Topic object derived from the supplied data
        @rtype: Topic
        @raises TypeError: any passed parameters have wrong types
        @raises ValueError: the passed topiclist_data is missing required data
        """
        if not isinstance(board, Board):
            raise TypeError('board must be a pywikibot.flow.Board object.')
        if not isinstance(root_uuid, UnicodeType):
            raise TypeError('Topic/root UUID must be a string.')

        topic = cls(board.site, 'Topic:' + root_uuid)
        topic._root = Post.fromJSON(topic, root_uuid, topiclist_data)
        topic._uuid = root_uuid
        return topic

    @property
    def root(self):
        """The root post of this topic."""
        if not hasattr(self, '_root'):
            self._root = Post.fromJSON(self, self.uuid, self._data)
        return self._root

    @property
    def is_locked(self):
        """Whether this topic is locked."""
        return self.root._current_revision['isLocked']

    @property
    def is_moderated(self):
        """Whether this topic is moderated."""
        return self.root._current_revision['isModerated']

    def replies(self, format='wikitext', force=False):
        """A list of replies to this topic's root post.

        @param format: Content format to return contents in
        @type format: str ('wikitext', 'html', or 'fixed-html')
        @param force: Whether to reload from the API instead of using the cache
        @type force: bool
        @return: The replies of this topic's root post
        @rtype: list of Posts
        """
        return self.root.replies(format=format, force=force)

    def reply(self, content, format='wikitext'):
        """A convenience method to reply to this topic's root post.

        @param content: The content of the new post
        @type content: str
        @param format: The format of the given content
        @type format: str ('wikitext' or 'html')
        @return: The new reply to this topic's root post
        @rtype: Post
        """
        return self.root.reply(content, format)

    # Moderation
    def lock(self, reason):
        """Lock this topic.

        @param reason: The reason for locking this topic
        @type reason: str
        """
        self.site.lock_topic(self, True, reason)
        self._reload()

    def unlock(self, reason):
        """Unlock this topic.

        @param reason: The reason for unlocking this topic
        @type reason: str
        """
        self.site.lock_topic(self, False, reason)
        self._reload()

    def delete_mod(self, reason):
        """Delete this topic through the Flow moderation system.

        @param reason: The reason for deleting this topic.
        @type reason: str
        """
        self.site.delete_topic(self, reason)
        self._reload()

    def hide(self, reason):
        """Hide this topic.

        @param reason: The reason for hiding this topic.
        @type reason: str
        """
        self.site.hide_topic(self, reason)
        self._reload()

    def suppress(self, reason):
        """Suppress this topic.

        @param reason: The reason for suppressing this topic.
        @type reason: str
        """
        self.site.suppress_topic(self, reason)
        self._reload()

    def restore(self, reason):
        """Restore this topic.

        @param reason: The reason for restoring this topic.
        @type reason: str
        """
        self.site.restore_topic(self, reason)
        self._reload()


# Flow non-page-like objects
class Post(object):

    """A post to a Flow discussion topic."""

    def __init__(self, page, uuid):
        """
        Initializer.

        @param page: Flow topic
        @type page: Topic
        @param uuid: UUID of a Flow post
        @type uuid: str

        @raises TypeError: incorrect types of parameters
        """
        if not isinstance(page, Topic):
            raise TypeError('Page must be a Topic object')
        if not page.exists():
            raise NoPage(page, 'Topic must exist: %s')
        if not isinstance(uuid, UnicodeType):
            raise TypeError('Post UUID must be a string')

        self._page = page
        self._uuid = uuid

        self._content = {}

    @classmethod
    def fromJSON(cls, page, post_uuid, data):
        """
        Create a Post object using the data returned from the API call.

        @param page: A Flow topic
        @type page: Topic
        @param post_uuid: The UUID of the post
        @type post_uuid: str
        @param data: The JSON data returned from the API
        @type data: dict

        @return: A Post object
        @raises TypeError: data is not a dict
        @raises ValueError: data is missing required entries
        """
        post = cls(page, post_uuid)
        post._set_data(data)

        return post

    def _set_data(self, data):
        """Set internal data and cache content.

        @param data: The data to store internally
        @type data: dict
        @raises TypeError: data is not a dict
        @raises ValueError: missing data entries or post/revision not found
        """
        if not isinstance(data, dict):
            raise TypeError('Illegal post data (must be a dictionary).')
        if ('posts' not in data) or ('revisions' not in data):
            raise ValueError('Illegal post data (missing required data).')
        if self.uuid not in data['posts']:
            raise ValueError('Post not found in supplied data.')

        current_revision_id = data['posts'][self.uuid][0]
        if current_revision_id not in data['revisions']:
            raise ValueError('Current revision of post'
                             'not found in supplied data.')

        self._current_revision = data['revisions'][current_revision_id]
        if 'content' in self._current_revision:
            content = self._current_revision.pop('content')
            assert isinstance(content, dict)
            assert isinstance(content['content'], UnicodeType)
            self._content[content['format']] = content['content']

    def _load(self, format='wikitext', load_from_topic=False):
        """Load and cache the Post's data using the given content format."""
        if load_from_topic:
            data = self.page._load(format=format, force=True)
        else:
            data = self.site.load_post_current_revision(self.page, self.uuid,
                                                        format)
        self._set_data(data)
        return self._current_revision

    @property
    def uuid(self):
        """Return the UUID of the post.

        @return: UUID of the post
        @rtype: str
        """
        return self._uuid

    @property
    def site(self):
        """Return the site associated with the post.

        @return: Site associated with the post
        @rtype: Site
        """
        return self._page.site

    @property
    def page(self):
        """Return the page associated with the post.

        @return: Page associated with the post
        @rtype: FlowPage
        """
        return self._page

    @property
    def is_moderated(self):
        """Whether this post is moderated."""
        if not hasattr(self, '_current_revision'):
            self._load()
        return self._current_revision['isModerated']

    @property
    def creator(self):
        """The creator of this post."""
        if not hasattr(self, '_current_revision'):
            self._load()
        if not hasattr(self, '_creator'):
            self._creator = User(self.site,
                                 self._current_revision['creator']['name'])
        return self._creator

    def get(self, format='wikitext', force=False):
        """Return the contents of the post in the given format.

        @param force: Whether to reload from the API instead of using the cache
        @type force: bool
        @param format: Content format to return contents in
        @type format: str
        @return: The contents of the post in the given content format
        @rtype: str
        """
        if format not in self._content or force:
            self._load(format)
        return self._content[format]

    def replies(self, format='wikitext', force=False):
        """Return this post's replies.

        @param format: Content format to return contents in
        @type format: str ('wikitext', 'html', or 'fixed-html')
        @param force: Whether to reload from the API instead of using the cache
        @type force: bool
        @return: This post's replies
        @rtype: list of Posts
        """
        if format not in ('wikitext', 'html', 'fixed-html'):
            raise ValueError('Invalid content format.')

        if hasattr(self, '_replies') and not force:
            return self._replies

        # load_from_topic workaround due to T106733
        # (replies not returned by view-post)
        if not hasattr(self, '_current_revision') or force:
            self._load(format, load_from_topic=True)

        reply_uuids = self._current_revision['replies']
        self._replies = [Post(self.page, uuid) for uuid in reply_uuids]

        return self._replies

    def reply(self, content, format='wikitext'):
        """Reply to this post.

        @param content: The content of the new post
        @type content: str
        @param format: The format of the given content
        @type format: str ('wikitext' or 'html')
        @return: The new reply post
        @rtype: Post
        """
        self._load()
        if self.page.is_locked:
            raise LockedPage(self.page, 'Topic %s is locked.')

        reply_url = self._current_revision['actions']['reply']['url']
        parsed_url = urlparse(reply_url)
        params = parse_qs(parsed_url.query)
        reply_to = params['topic_postId']
        if self.uuid == reply_to:
            del self._current_revision
            del self._replies
        data = self.site.reply_to_post(self.page, reply_to, content, format)
        post = Post(self.page, data['post-id'])
        return post

    # Moderation
    def delete(self, reason):
        """Delete this post through the Flow moderation system.

        @param reason: The reason for deleting this post.
        @type reason: str
        """
        self.site.delete_post(self, reason)
        self._load()

    def hide(self, reason):
        """Hide this post.

        @param reason: The reason for hiding this post.
        @type reason: str
        """
        self.site.hide_post(self, reason)
        self._load()

    def suppress(self, reason):
        """Suppress this post.

        @param reason: The reason for suppressing this post.
        @type reason: str
        """
        self.site.suppress_post(self, reason)
        self._load()

    def restore(self, reason):
        """Restore this post.

        @param reason: The reason for restoring this post.
        @type reason: str
        """
        self.site.restore_post(self, reason)
        self._load()

    def thank(self):
        """Thank the user who made this post."""
        self.site.thank_post(self)



# -*- coding: utf-8 -*-
"""Module with the Graphviz drawing calls."""
#
# (C) Pywikibot team, 2006-2018
#
# Distributed under the terms of the MIT license.
#
from __future__ import absolute_import, division, unicode_literals

from collections import Counter
import itertools
import threading

try:
    import pydot
except ImportError as e:
    pydot = e

import pywikibot

from pywikibot import config2 as config

# deprecated value
pydotfound = not isinstance(pydot, ImportError)


class GraphImpossible(Exception):

    """Drawing a graph is not possible on your system."""


class GraphSavingThread(threading.Thread):

    """
    Threaded graph renderer.

    Rendering a graph can take extremely long. We use
    multithreading because of that.

    TODO: Find out if several threads running in parallel
    can slow down the system too much. Consider adding a
    mechanism to kill a thread if it takes too long.
    """

    def __init__(self, graph, originPage):
        """Initializer."""
        threading.Thread.__init__(self)
        self.graph = graph
        self.originPage = originPage

    def run(self):
        """Write graphs to the data directory."""
        for format in config.interwiki_graph_formats:
            filename = config.datafilepath(
                'interwiki-graphs/' + getFilename(self.originPage, format))
            if self.graph.write(filename, prog='dot', format=format):
                pywikibot.output('Graph saved as %s' % filename)
            else:
                pywikibot.output('Graph could not be saved as %s' % filename)


class Subject(object):

    """Data about a page with translations on multiple wikis."""

    def __init__(self, origin=None):
        """Initializer.

        @param originPage: the page on the 'origin' wiki
        @type originPage: pywikibot.page.Page
        """
        # Remember the "origin page"
        self._origin = origin

        # Temporary variable to support git blame; do not use
        originPage = origin

        self.found_in = None

        # foundIn is a dictionary where pages are keys and lists of
        # pages are values. It stores where we found each page.
        # As we haven't yet found a page that links to the origin page, we
        # start with an empty list for it.
        if originPage:
            self.foundIn = {self.originPage: []}
        else:
            self.foundIn = {}

    @property
    def origin(self):
        """Page on the origin wiki."""
        return self._origin

    @origin.setter
    def origin(self, value):
        self._origin = value

    @property
    def originPage(self):
        """Deprecated property for the origin page.

        DEPRECATED. Use origin.
        """
        # TODO: deprecate this property
        return self.origin

    @originPage.setter
    def originPage(self, value):
        self.origin = value

    @property
    def foundIn(self):
        """Mapping of pages to others pages interwiki linked to it.

        DEPRECATED. Use found_in.
        """
        # TODO: deprecate this property
        return self.found_in

    @foundIn.setter
    def foundIn(self, value):
        """Temporary property setter to support code migration."""
        self.found_in = value


class GraphDrawer(object):

    """Graphviz (dot) code creator."""

    def __init__(self, subject):
        """Initializer.

        @param subject: page data to graph
        @type subject: pywikibot.interwiki_graph.Subject

        @raises GraphImpossible: pydot is not installed
        """
        if isinstance(pydot, ImportError):
            raise GraphImpossible('pydot is not installed: %s.' % pydot)
        self.graph = None
        self.subject = subject

    def getLabel(self, page):
        """Get label for page."""
        return '"%s:%s"' % (page.site.code, page.title())

    def _octagon_site_set(self):
        """Build a list of sites with more than one valid page."""
        page_list = self.subject.found_in.keys()

        # Only track sites of normal pages
        each_site = [page.site for page in page_list
                     if page.exists() and not page.isRedirectPage()]

        return {x[0] for x in itertools.takewhile(
            lambda x: x[1] > 1,
            Counter(each_site).most_common())}

    def addNode(self, page):
        """Add a node for page."""
        node = pydot.Node(self.getLabel(page), shape='rectangle')
        node.set_URL('"http://%s%s"'
                     % (page.site.hostname(),
                        page.site.get_address(page.title(as_url=True))))
        node.set_style('filled')
        node.set_fillcolor('white')
        node.set_fontsize('11')
        if not page.exists():
            node.set_fillcolor('red')
        elif page.isRedirectPage():
            node.set_fillcolor('blue')
        elif page.isDisambig():
            node.set_fillcolor('orange')
        if page.namespace() != self.subject.originPage.namespace():
            node.set_color('green')
            node.set_style('filled,bold')
        if page.site in self.octagon_sites:
            # mark conflict by octagonal node
            node.set_shape('octagon')
        self.graph.add_node(node)

    def addDirectedEdge(self, page, refPage):
        """Add a directed edge from refPage to page."""
        # if page was given as a hint, referrers would be [None]
        if refPage is not None:
            sourceLabel = self.getLabel(refPage)
            targetLabel = self.getLabel(page)
            edge = pydot.Edge(sourceLabel, targetLabel)

            oppositeEdge = self.graph.get_edge(targetLabel, sourceLabel)
            if oppositeEdge:
                if isinstance(oppositeEdge, list):
                    # bugfix for pydot >= 1.0.3
                    oppositeEdge = oppositeEdge[0]
                oppositeEdge.set_dir('both')
            # workaround for sf.net bug 401: prevent duplicate edges
            # (it is unclear why duplicate edges occur)
            # https://sourceforge.net/p/pywikipediabot/bugs/401/
            elif self.graph.get_edge(sourceLabel, targetLabel):
                pywikibot.output(
                    'BUG: Tried to create duplicate edge from %s to %s'
                    % (refPage.title(as_link=True), page.title(as_link=True)))
                # duplicate edges would be bad because then get_edge() would
                # give a list of edges, not a single edge when we handle the
                # opposite edge.
            else:
                # add edge
                if refPage.site == page.site:
                    edge.set_color('blue')
                elif not page.exists():
                    # mark dead links
                    edge.set_color('red')
                elif refPage.isDisambig() != page.isDisambig():
                    # mark links between disambiguation and non-disambiguation
                    # pages
                    edge.set_color('orange')
                if refPage.namespace() != page.namespace():
                    edge.set_color('green')
                self.graph.add_edge(edge)

    def saveGraphFile(self):
        """Write graphs to the data directory."""
        thread = GraphSavingThread(self.graph, self.subject.originPage)
        thread.start()

    def createGraph(self):
        """
        Create graph of the interwiki links.

        For more info see U{https://meta.wikimedia.org/wiki/Interwiki_graphs}
        """
        pywikibot.output('Preparing graph for %s'
                         % self.subject.originPage.title())
        # create empty graph
        self.graph = pydot.Dot()
        # self.graph.set('concentrate', 'true')

        self.octagon_sites = self._octagon_site_set()

        for page in self.subject.foundIn.keys():
            # a node for each found page
            self.addNode(page)
        # mark start node by pointing there from a black dot.
        firstLabel = self.getLabel(self.subject.originPage)
        self.graph.add_node(pydot.Node('start', shape='point'))
        self.graph.add_edge(pydot.Edge('start', firstLabel))
        for page, referrers in self.subject.foundIn.items():
            for refPage in referrers:
                self.addDirectedEdge(page, refPage)
        self.saveGraphFile()


def getFilename(page, extension=None):
    """
    Create a filename that is unique for the page.

    @param page: page used to create the new filename
    @type page: pywikibot.page.Page
    @param extension: file extension
    @type extension: str
    @return: filename of <family>-<lang>-<page>.<ext>
    @rtype: str
    """
    filename = '%s-%s-%s' % (page.site.family.name,
                             page.site.code,
                             page.title(as_filename=True))
    if extension:
        filename += '.%s' % extension
    return filename



# -*- coding: utf-8 -*-
"""Family module for OpenStreetMap wiki."""
#
# (C) Pywikibot team, 2009-2018
#
# Distributed under the terms of the MIT license.
#
from __future__ import absolute_import, division, unicode_literals

from pywikibot import family


# The project wiki of OpenStreetMap (OSM).
class Family(family.SingleSiteFamily):

    """Family class for OpenStreetMap wiki."""

    name = 'osm'
    domain = 'wiki.openstreetmap.org'
    code = 'en'

    def protocol(self, code):
        """Return https as the protocol for this family."""
        return 'https'



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Text editor class for your favourite editor."""
#
# (C) Pywikibot team, 2004-2020
#
# Distributed under the terms of the MIT license.
#
from __future__ import absolute_import, division, unicode_literals

import codecs
import os
import subprocess
import tempfile

from sys import platform

import pywikibot

from pywikibot import config
from pywikibot.tools import deprecated

try:
    from pywikibot.userinterfaces import gui
except ImportError as e:
    gui = e


class TextEditor(object):

    """Text editor."""

    def _command(self, file_name, text, jump_index=None):
        """Return editor selected in user-config.py."""
        if jump_index:
            # Some editors make it possible to mark occurrences of substrings,
            # or to jump to the line of the first occurrence.
            # TODO: Find a better solution than hardcoding these, e.g. a config
            # option.
            line = text[:jump_index].count('\n')
            column = jump_index - (text[:jump_index].rfind('\n') + 1)
        else:
            line = column = 0
        # Linux editors. We use startswith() because some users might use
        # parameters.
        if config.editor.startswith('kate'):
            command = ['-l', '%i' % (line + 1), '-c', '%i' % (column + 1)]
        elif config.editor.startswith('gedit'):
            command = ['+%i' % (line + 1)]  # seems not to support columns
        elif config.editor.startswith('emacs'):
            command = ['+%i' % (line + 1)]  # seems not to support columns
        elif config.editor.startswith('jedit'):
            command = ['+line:%i' % (line + 1)]  # seems not to support columns
        elif config.editor.startswith('vim'):
            command = ['+%i' % (line + 1)]  # seems not to support columns
        elif config.editor.startswith('nano'):
            command = ['+%i,%i' % (line + 1, column + 1)]
        # Windows editors
        elif config.editor.lower().endswith('notepad++.exe'):
            command = ['-n%i' % (line + 1)]  # seems not to support columns
        else:
            command = []

        # See T102465 for problems relating to using config.editor unparsed.
        command = [config.editor] + command + [file_name]
        pywikibot.log('Running editor: %s' % TextEditor._concat(command))
        return command

    @staticmethod
    def _concat(command):
        return ' '.join("'{0}'".format(part) if ' ' in part else part
                        for part in command)

    @deprecated('_command (should not be used from the outside)',
                since='20150111')
    def command(self, tempFilename, text, jumpIndex=None):
        """Return editor selected in user-config.py."""
        return TextEditor._concat(self._command(tempFilename, text, jumpIndex))

    def edit(self, text, jumpIndex=None, highlight=None):
        """
        Call the editor and thus allows the user to change the text.

        Halts the thread's operation until the editor is closed.

        @param text: the text to be edited
        @type text: str
        @param jumpIndex: position at which to put the caret
        @type jumpIndex: int
        @param highlight: each occurrence of this substring will be highlighted
        @type highlight: str
        @return: the modified text, or None if the user didn't save the text
            file in his text editor
        @rtype: str or None
        """
        if config.editor:
            handle, tempFilename = tempfile.mkstemp()
            tempFilename = '%s.%s' % (tempFilename,
                                      config.editor_filename_extension)
            try:
                with codecs.open(tempFilename, 'w',
                                 encoding=config.editor_encoding) as tempFile:
                    tempFile.write(text)
                creationDate = os.stat(tempFilename).st_mtime
                cmd = self._command(tempFilename, text, jumpIndex)
                subprocess.call(cmd, shell=platform == 'win32')
                lastChangeDate = os.stat(tempFilename).st_mtime
                if lastChangeDate == creationDate:
                    # Nothing changed
                    return None
                else:
                    with codecs.open(
                        tempFilename, 'r', encoding=config.editor_encoding
                    ) as temp_file:
                        newcontent = temp_file.read()
                    return newcontent
            finally:
                os.close(handle)
                os.unlink(tempFilename)

        if isinstance(gui, ImportError):
            raise ImportError(
                'Could not load GUI modules: %s\nNo editor available.\n'
                'Set your favourite editor in user-config.py "editor", '
                'or install python packages tkinter and idlelib, which '
                'are typically part of Python but may be packaged separately '
                'on your platform.\n' % gui)

        return pywikibot.ui.editText(
            text, jumpIndex=jumpIndex, highlight=highlight)



# -*- coding: utf-8 -*-
"""Family module for Wikimedia Commons."""
#
# (C) Pywikibot team, 2005-2018
#
# Distributed under the terms of the MIT license.
#
from __future__ import absolute_import, division, unicode_literals

from pywikibot import family


# The Wikimedia Commons family
class Family(family.WikimediaFamily):

    """Family class for Wikimedia Commons."""

    name = 'commons'

    langs = {
        'commons': 'commons.wikimedia.org',
        'test': 'test-commons.wikimedia.org',
        'beta': 'commons.wikimedia.beta.wmflabs.org'
    }

    interwiki_forward = 'wikipedia'

    # Templates that indicate a category redirect
    # Redirects to these templates are automatically included
    category_redirect_templates = {
        '_default': (
            'Category redirect',
            'Synonym taxon category redirect',
            'Invalid taxon category redirect',
            'Monotypic taxon category redirect',
            'Endashcatredirect',
        ),
    }

    # Subpages for documentation.
    doc_subpages = {
        '_default': (('/doc', ), ['commons']),
    }



# -*- coding: utf-8 -*-
"""
XML reading module.

Each XmlEntry object represents a page, as read from an XML source

The XmlDump class reads a pages_current XML dump (like the ones offered on
https://dumps.wikimedia.org/backup-index.html) and offers a generator over
XmlEntry objects which can be used by other bots.
"""
#
# (C) Pywikibot team, 2005-2020
#
# Distributed under the terms of the MIT license.
#
from __future__ import absolute_import, division, unicode_literals

import re
import threading

from xml.etree.ElementTree import iterparse

import xml.sax

from pywikibot.tools import open_archive


def parseRestrictions(restrictions):
    """
    Parse the characters within a restrictions tag.

    Returns strings representing user groups allowed to edit and
    to move a page, where None means there are no restrictions.
    """
    if not restrictions:
        return None, None
    editRestriction = None
    moveRestriction = None
    editLockMatch = re.search('edit=([^:]*)', restrictions)
    if editLockMatch:
        editRestriction = editLockMatch.group(1)
    moveLockMatch = re.search('move=([^:]*)', restrictions)
    if moveLockMatch:
        moveRestriction = moveLockMatch.group(1)
    if restrictions == 'sysop':
        editRestriction = 'sysop'
        moveRestriction = 'sysop'
    return editRestriction, moveRestriction


class XmlEntry(object):

    """Represent a page."""

    def __init__(self, title, ns, id, text, username, ipedit, timestamp,
                 editRestriction, moveRestriction, revisionid, comment,
                 redirect):
        """Initializer."""
        # TODO: there are more tags we can read.
        self.title = title
        self.ns = ns
        self.id = id
        self.text = text
        self.username = username.strip()
        self.ipedit = ipedit
        self.timestamp = timestamp
        self.editRestriction = editRestriction
        self.moveRestriction = moveRestriction
        self.revisionid = revisionid
        self.comment = comment
        self.isredirect = redirect


class XmlParserThread(threading.Thread):

    """
    XML parser that will run as a single thread.

    This allows the XmlDump
    generator to yield pages before the parser has finished reading the
    entire dump.

    There surely are more elegant ways to do this.
    """

    def __init__(self, filename, handler):
        """Initializer."""
        threading.Thread.__init__(self)
        self.filename = filename
        self.handler = handler

    def run(self):
        """Parse the file in a single thread."""
        xml.sax.parse(self.filename, self.handler)


class XmlDump(object):

    """
    Represents an XML dump file.

    Reads the local file at initialization,
    parses it, and offers access to the resulting XmlEntries via a generator.

    @param allrevisions: boolean
        If True, parse all revisions instead of only the latest one.
        Default: False.
    """

    def __init__(self, filename, allrevisions=False):
        """Initializer."""
        self.filename = filename
        if allrevisions:
            self._parse = self._parse_all
        else:
            self._parse = self._parse_only_latest

    def parse(self):
        """Generator using ElementTree iterparse function."""
        with open_archive(self.filename) as source:
            # iterparse's event must be a str but they are unicode with
            # unicode_literals in Python 2
            context = iterparse(source, events=(str('start'), str('end'),
                                                str('start-ns')))
            self.root = None

            for event, elem in context:
                if event == 'start-ns' and elem[0] == '':
                    self.uri = elem[1]
                    continue
                if event == 'start' and self.root is None:
                    self.root = elem
                    continue
                for rev in self._parse(event, elem):
                    yield rev

    def _parse_only_latest(self, event, elem):
        """Parser that yields only the latest revision."""
        if event == 'end' and elem.tag == '{%s}page' % self.uri:
            self._headers(elem)
            revision = elem.find('{%s}revision' % self.uri)
            yield self._create_revision(revision)
            elem.clear()
            self.root.clear()

    def _parse_all(self, event, elem):
        """Parser that yields all revisions."""
        if event == 'start' and elem.tag == '{%s}page' % self.uri:
            self._headers(elem)
        if event == 'end' and elem.tag == '{%s}revision' % self.uri:
            yield self._create_revision(elem)
            elem.clear()
            self.root.clear()

    def _headers(self, elem):
        """Extract headers from XML chunk."""
        self.title = elem.findtext('{%s}title' % self.uri)
        self.ns = elem.findtext('{%s}ns' % self.uri)
        self.pageid = elem.findtext('{%s}id' % self.uri)
        self.restrictions = elem.findtext('{%s}restrictions' % self.uri)
        self.isredirect = elem.findtext('{%s}redirect' % self.uri) is not None
        self.editRestriction, self.moveRestriction = parseRestrictions(
            self.restrictions)

    def _create_revision(self, revision):
        """Create a Single revision."""
        revisionid = revision.findtext('{%s}id' % self.uri)
        timestamp = revision.findtext('{%s}timestamp' % self.uri)
        comment = revision.findtext('{%s}comment' % self.uri)
        contributor = revision.find('{%s}contributor' % self.uri)
        ipeditor = contributor.findtext('{%s}ip' % self.uri)
        username = ipeditor or contributor.findtext('{%s}username' % self.uri)
        # could get comment, minor as well
        text = revision.findtext('{%s}text' % self.uri)
        return XmlEntry(title=self.title,
                        ns=self.ns,
                        id=self.pageid,
                        text=text or '',
                        username=username or '',  # username might be deleted
                        ipedit=bool(ipeditor),
                        timestamp=timestamp,
                        editRestriction=self.editRestriction,
                        moveRestriction=self.moveRestriction,
                        revisionid=revisionid,
                        comment=comment,
                        redirect=self.isredirect
                        )



# -*- coding: utf-8 -*-
"""Diff module."""
#
# (C) Pywikibot team, 2014-2020
#
# Distributed under the terms of the MIT license.
#
from __future__ import absolute_import, division, unicode_literals

import difflib
import math

try:
    from collections.abc import Sequence
except ImportError:  # Python 2.7
    from collections import Sequence
from difflib import _format_range_unified as format_range_unified

import pywikibot
from pywikibot.tools import chars, deprecated_args, PY2
from pywikibot.tools.formatter import color_format

if not PY2:
    from itertools import zip_longest
else:
    from itertools import izip_longest as zip_longest


class Hunk(object):

    """One change hunk between a and b.

    Note: parts of this code are taken from by difflib.get_grouped_opcodes().

    """

    APPR = 1
    NOT_APPR = -1
    PENDING = 0

    def __init__(self, a, b, grouped_opcode):
        """
        Initializer.

        @param a: sequence of lines
        @param b: sequence of lines
        @param grouped_opcode: list of 5-tuples describing how to turn a into
            b. It has the same format as returned by difflib.get_opcodes().
        """
        self.a = a
        self.b = b
        self.group = grouped_opcode
        self.header = ''
        self.colors = {
            '+': 'lightgreen',
            '-': 'lightred',
        }
        self.bg_colors = {
            '+': 'lightgreen',
            '-': 'lightred',
        }

        self.diff = list(self.create_diff())
        self.diff_plain_text = ''.join(self.diff)
        self.diff_text = ''.join(self.format_diff())

        first, last = self.group[0], self.group[-1]
        self.a_rng = (first[1], last[2])
        self.b_rng = (first[3], last[4])

        self.header = self.get_header()
        self.diff_plain_text = '%s\n%s' % (self.header, self.diff_plain_text)
        self.diff_text = '%s' % self.diff_text

        self.reviewed = self.PENDING

    def get_header(self):
        """Provide header of unified diff."""
        return self.get_header_text(self.a_rng, self.b_rng) + '\n'

    @staticmethod
    def get_header_text(a_rng, b_rng, affix='@@'):
        """Provide header for any ranges."""
        a_rng = format_range_unified(*a_rng)
        b_rng = format_range_unified(*b_rng)
        return '{0} -{1} +{2} {0}'.format(affix, a_rng, b_rng)

    def create_diff(self):
        """Generator of diff text for this hunk, without formatting."""
        # make sure each line ends with '\n' to prevent
        # behaviour like http://bugs.python.org/issue2142
        def check_line(line):
            return line if line.endswith('\n') else line + '\n'

        for tag, i1, i2, j1, j2 in self.group:
            # equal/delete/insert add additional space after the sign as it's
            # what difflib.ndiff does do too.
            if tag == 'equal':
                for line in self.a[i1:i2]:
                    yield '  ' + check_line(line)
            elif tag == 'delete':
                for line in self.a[i1:i2]:
                    yield '- ' + check_line(line)
            elif tag == 'insert':
                for line in self.b[j1:j2]:
                    yield '+ ' + check_line(line)
            elif tag == 'replace':
                for line in difflib.ndiff(self.a[i1:i2], self.b[j1:j2]):
                    yield check_line(line)

    def format_diff(self):
        """Color diff lines."""
        diff = iter(self.diff)

        fmt, line1, line2 = '', '', next(diff)
        for line in diff:
            fmt, line1, line2 = line1, line2, line
            # do not show lines starting with '?'.
            if line1.startswith('?'):
                continue
            if line2.startswith('?'):
                yield self.color_line(line1, line2)
                # do not try to reuse line2 as format at next iteration
                # if already used for an added line.
                if line1.startswith('+'):
                    line2 = ''
                continue
            if line1.startswith('-'):
                # Color whole line to be removed.
                yield self.color_line(line1)
            elif line1.startswith('+'):
                # Reuse last available fmt as diff line, if possible,
                # or color whole line to be added.
                fmt = fmt if fmt.startswith('?') else ''
                fmt = fmt[:min(len(fmt), len(line1))]
                fmt = fmt if fmt else None
                yield self.color_line(line1, fmt)

        # handle last line
        # If line line2 is removed, color the whole line.
        # If line line2 is added, check if line1 is a '?-type' line, to prevent
        # the entire line line2 to be colored (see T130572).
        # The case where line2 start with '?' has been covered already.
        if line2.startswith('-'):
            # Color whole line to be removed.
            yield self.color_line(line2)
        elif line2.startswith('+'):
            # Reuse last available line1 as diff line, if possible,
            # or color whole line to be added.
            fmt = line1 if line1.startswith('?') else ''
            fmt = fmt[:min(len(fmt), len(line2))]
            fmt = fmt if fmt else None
            yield self.color_line(line2, fmt)

    def color_line(self, line, line_ref=None):
        """Color line characters.

        If line_ref is None, the whole line is colored.
        If line_ref[i] is not blank, line[i] is colored.
        Color depends if line starts with +/-.

        line: string
        line_ref: string.

        """
        color = line[0]

        if line_ref is None:
            if color in self.colors:
                colored_line = color_format('{color}{0}{default}',
                                            line, color=self.colors[color])
                return colored_line
            else:
                return line

        colored_line = ''
        color_closed = True
        for char, char_ref in zip_longest(
            line, line_ref.strip(), fillvalue=' '
        ):
            char_tagged = char
            if color_closed:
                if char_ref != ' ':
                    if char != ' ':
                        apply_color = self.colors[color]
                    else:
                        apply_color = 'default;' + self.bg_colors[color]
                    char_tagged = color_format('{color}{0}',
                                               char, color=apply_color)
                    color_closed = False
            else:
                if char_ref == ' ':
                    char_tagged = color_format('{default}{0}', char)
                    color_closed = True
            colored_line += char_tagged

        if not color_closed:
            colored_line += color_format('{default}')

        return colored_line

    def apply(self):
        """Turn a into b for this hunk."""
        return self.b[self.b_rng[0]:self.b_rng[1]]

    def __str__(self):
        """Return the diff as plain text."""
        return ''.join(self.diff_plain_text)

    def __repr__(self):
        """Return a reconstructable representation."""
        # TODO
        return '%s(a, b, %s)' \
               % (self.__class__.__name__, self.group)


class _SuperHunk(Sequence):

    def __init__(self, hunks):
        self._hunks = hunks
        self.a_rng = (self._hunks[0].a_rng[0], self._hunks[-1].a_rng[1])
        self.b_rng = (self._hunks[0].b_rng[0], self._hunks[-1].b_rng[1])
        self.pre_context = self._hunks[0].pre_context
        self.post_context = self._hunks[0].post_context

    def __getitem__(self, idx):
        return self._hunks[idx]

    def __len__(self):
        return len(self._hunks)

    def split(self):
        return [_SuperHunk([hunk]) for hunk in self._hunks]

    @property
    def reviewed(self):
        assert len({hunk.reviewed for hunk in self._hunks}) == 1, \
            'All hunks should have the same review status'
        return self._hunks[0].reviewed

    @reviewed.setter
    def reviewed(self, reviewed):
        for hunk in self._hunks:
            hunk.reviewed = reviewed


class PatchManager(object):

    """Apply patches to text_a to obtain a new text.

    If all hunks are approved, text_b will be obtained.
    """

    @deprecated_args(n='context')
    def __init__(self, text_a, text_b, context=0, by_letter=False,
                 replace_invisible=False):
        """Initializer.

        @param text_a: base text
        @type text_a: basestring
        @param text_b: target text
        @type text_b: basestring
        @param context: number of lines which are context
        @type context: int
        @param by_letter: if text_a and text_b are single lines, comparison can
            be done letter by letter.
        @type by_letter: bool
        @param replace_invisible: Replace invisible characters like U+200e with
            the charnumber in brackets (e.g. <200e>).
        @type replace_invisible: bool
        """
        if '\n' in text_a or '\n' in text_b or not by_letter:
            self.a = text_a.splitlines(1)
            self.b = text_b.splitlines(1)
        else:
            self.a = text_a
            self.b = text_b

        # groups and hunk have same order (one hunk correspond to one group).
        s = difflib.SequenceMatcher(None, self.a, self.b)
        self.groups = list(s.get_grouped_opcodes(0))
        self.hunks = []
        previous_hunk = None
        for group in self.groups:
            hunk = Hunk(self.a, self.b, group)
            self.hunks.append(hunk)
            hunk.pre_context = hunk.a_rng[0]
            if previous_hunk:
                hunk.pre_context -= previous_hunk.a_rng[1]
                previous_hunk.post_context = hunk.pre_context
            previous_hunk = hunk
        if self.hunks:
            self.hunks[-1].post_context = len(self.a) - self.hunks[-1].a_rng[1]
        # blocks are a superset of hunk, as include also parts not
        # included in any hunk.
        self.blocks = self.get_blocks()
        self.context = context
        self._super_hunks = self._generate_super_hunks()
        self._replace_invisible = replace_invisible

    def get_blocks(self):
        """Return list with blocks of indexes.

        Format of each block::

            [-1, (i1, i2), (-1, -1)] -> block a[i1:i2] does not change from
                a to b then is there is no corresponding hunk.
            [hunk index, (i1, i2), (j1, j2)] -> block a[i1:i2] becomes b[j1:j2]
        """
        blocks = []
        i2 = 0
        for hunk_idx, group in enumerate(self.groups):

            first, last = group[0], group[-1]
            i1, prev_i2, i2 = first[1], i2, last[2]

            # there is a section of unchanged text before this hunk.
            if prev_i2 < i1:
                rng = (-1, (prev_i2, i1), (-1, -1))
                blocks.append(rng)

            rng = (hunk_idx, (first[1], last[2]), (first[3], last[4]))
            blocks.append(rng)

        # there is a section of unchanged text at the end of a, b.
        if i2 < len(self.a):
            rng = (-1, (i2, len(self.a)), (-1, -1))
            blocks.append(rng)

        return blocks

    def print_hunks(self):
        """Print the headers and diff texts of all hunks to the output."""
        if self.hunks:
            pywikibot.output('\n'.join(self._generate_diff(super_hunk)
                                       for super_hunk in self._super_hunks))

    def _generate_super_hunks(self, hunks=None):
        if hunks is None:
            hunks = self.hunks
        if self.context:
            # Determine if two hunks are connected by self.context
            super_hunk = []
            super_hunks = [super_hunk]
            for hunk in hunks:
                # self.context * 2, because if self.context is 2 the hunks
                # would be directly adjacent when 4 lines in between and for
                # anything below 4 they share lines.
                # not super_hunk == first hunk as any other super_hunk is
                # created with one hunk
                if (not super_hunk or hunk.pre_context <= self.context * 2):
                    # previous hunk has shared/adjacent self.context lines
                    super_hunk += [hunk]
                else:
                    super_hunk = [hunk]
                    super_hunks += [super_hunk]
        else:
            super_hunks = [[hunk] for hunk in hunks]
        return [_SuperHunk(sh) for sh in super_hunks]

    def _get_context_range(self, super_hunk):
        """Dynamically determine context range for a super hunk."""
        a0, a1 = super_hunk.a_rng
        b0, b1 = super_hunk.b_rng
        return ((a0 - min(super_hunk.pre_context, self.context),
                 a1 + min(super_hunk.post_context, self.context)),
                (b0 - min(super_hunk.pre_context, self.context),
                 b1 + min(super_hunk.post_context, self.context)))

    def _generate_diff(self, hunks):
        """Generate a diff text for the given hunks."""
        def extend_context(start, end):
            """Add context lines."""
            return ''.join('  {0}\n'.format(line.rstrip())
                           for line in self.a[start:end])

        context_range = self._get_context_range(hunks)

        output = color_format('{aqua}{0}{default}\n{1}',
                              Hunk.get_header_text(*context_range),
                              extend_context(context_range[0][0],
                                             hunks[0].a_rng[0]))
        previous_hunk = None
        for hunk in hunks:
            if previous_hunk:
                output += extend_context(previous_hunk.a_rng[1], hunk.a_rng[0])
            previous_hunk = hunk
            output += hunk.diff_text
        output += extend_context(hunks[-1].a_rng[1], context_range[0][1])
        if self._replace_invisible:
            output = chars.replace_invisible(output)
        return output

    def review_hunks(self):
        """Review hunks."""
        def find_pending(start, end):
            step = -1 if start > end else +1
            for pending in range(start, end, step):
                if super_hunks[pending].reviewed == Hunk.PENDING:
                    return pending

        # TODO: Missing commands (compared to git --patch): edit and search
        help_msg = {'y': 'accept this hunk',
                    'n': 'do not accept this hunk',
                    'q': 'do not accept this hunk and quit reviewing',
                    'a': 'accept this hunk and all other pending',
                    'd': 'do not apply this hunk or any of the later hunks in '
                         'the file',
                    'g': 'select a hunk to go to',
                    'j': 'leave this hunk undecided, see next undecided hunk',
                    'J': 'leave this hunk undecided, see next hunk',
                    'k': 'leave this hunk undecided, see previous undecided '
                         'hunk',
                    'K': 'leave this hunk undecided, see previous hunk',
                    's': 'split this hunk into smaller ones',
                    '?': 'help',
                    }

        super_hunks = self._generate_super_hunks(
            h for h in self.hunks if h.reviewed == Hunk.PENDING)
        position = 0

        while any(any(hunk.reviewed == Hunk.PENDING for hunk in super_hunk)
                  for super_hunk in super_hunks):

            super_hunk = super_hunks[position]

            next_pending = find_pending(position + 1, len(super_hunks))
            prev_pending = find_pending(position - 1, -1)

            answers = ['y', 'n', 'q', 'a', 'd', 'g']
            if next_pending is not None:
                answers += ['j']
            if position < len(super_hunks) - 1:
                answers += ['J']
            if prev_pending is not None:
                answers += ['k']
            if position > 0:
                answers += ['K']
            if len(super_hunk) > 1:
                answers += ['s']
            answers += ['?']

            pywikibot.output(self._generate_diff(super_hunk))
            choice = pywikibot.input('Accept this hunk [{0}]?'.format(
                ','.join(answers)))
            if choice not in answers:
                choice = '?'

            if choice in ['y', 'n']:
                super_hunk.reviewed = \
                    Hunk.APPR if choice == 'y' else Hunk.NOT_APPR
                if next_pending is not None:
                    position = next_pending
                else:
                    position = find_pending(0, position)
            elif choice == 'q':
                for super_hunk in super_hunks:
                    for hunk in super_hunk:
                        if hunk.reviewed == Hunk.PENDING:
                            hunk.reviewed = Hunk.NOT_APPR
            elif choice in ['a', 'd']:
                for super_hunk in super_hunks[position:]:
                    for hunk in super_hunk:
                        if hunk.reviewed == Hunk.PENDING:
                            hunk.reviewed = \
                                Hunk.APPR if choice == 'a' else Hunk.NOT_APPR
                position = find_pending(0, position)
            elif choice == 'g':
                hunk_list = []
                rng_width = 18
                for index, super_hunk in enumerate(super_hunks, start=1):
                    assert -1 <= super_hunk.reviewed <= 1, \
                        "The super hunk's review status is unknown."
                    status = ' +-'[super_hunk.reviewed]

                    if super_hunk[0].a_rng[1] - super_hunk[0].a_rng[0] > 0:
                        mode = '-'
                        first = self.a[super_hunk[0].a_rng[0]]
                    else:
                        mode = '+'
                        first = self.b[super_hunk[0].b_rng[0]]
                    hunk_list += [(status, index,
                                   Hunk.get_header_text(
                                       *self._get_context_range(super_hunk),
                                       affix=''),
                                   mode, first)]
                    rng_width = max(len(hunk_list[-1][2]), rng_width)
                line_template = ('{0}{1} {2: >'
                                 + str(int(math.log10(len(super_hunks)) + 1))
                                 + '}: {3: <' + str(rng_width) + '} {4}{5}')
                # the last entry is the first changed line which usually ends
                # with a \n (only the last may not, which is covered by the
                # if-condition following this block)
                hunk_list = ''.join(
                    line_template.format(
                        '*' if hunk_entry[1] == position + 1 else
                        ' ', *hunk_entry)
                    for hunk_entry in hunk_list)
                if hunk_list.endswith('\n'):
                    hunk_list = hunk_list[:-1]
                pywikibot.output(hunk_list)
                next_hunk = pywikibot.input('Go to which hunk?')
                try:
                    next_hunk_position = int(next_hunk) - 1
                except ValueError:
                    next_hunk_position = False
                if (next_hunk_position is not False
                        and 0 <= next_hunk_position < len(super_hunks)):
                    position = next_hunk_position
                elif next_hunk:  # nothing entered is silently ignored
                    pywikibot.error(
                        'Invalid hunk number "{0}"'.format(next_hunk))
            elif choice == 'j':
                position = next_pending
            elif choice == 'J':
                position += 1
            elif choice == 'k':
                position = prev_pending
            elif choice == 'K':
                position -= 1
            elif choice == 's':
                super_hunks = (super_hunks[:position]
                               + super_hunks[position].split()
                               + super_hunks[position + 1:])
                pywikibot.output(
                    'Split into {0} hunks'.format(len(super_hunk._hunks)))
            else:  # choice == '?':
                pywikibot.output(color_format(
                    '{purple}{0}{default}', '\n'.join(
                        '{0} -> {1}'.format(answer, help_msg[answer])
                        for answer in answers)))

    def apply(self):
        """Apply changes. If there are undecided changes, ask to review."""
        if any(h.reviewed == h.PENDING for h in self.hunks):
            pywikibot.output('There are unreviewed hunks.\n'
                             'Please review them before proceeding.\n')
            self.review_hunks()

        l_text = []
        for hunk_idx, (i1, i2), (j1, j2) in self.blocks:
            # unchanged text.
            if hunk_idx < 0:
                l_text.extend(self.a[i1:i2])
            # changed text; check if hunk is approved.
            else:
                hunk = self.hunks[hunk_idx]
                if hunk.reviewed == hunk.APPR:
                    l_text.extend(self.b[j1:j2])
                else:
                    l_text.extend(self.a[i1:i2])

        # Make a sanity check in case all are approved.
        if all(h.reviewed == h.APPR for h in self.hunks):
            assert ''.join(l_text) == ''.join(self.b)
        return l_text


def cherry_pick(oldtext, newtext, n=0, by_letter=False):
    """Propose a list of changes for approval.

    Text with approved changes will be returned.
    n: int, line of context as defined in difflib.get_grouped_opcodes().
    by_letter: if text_a and text_b are single lines, comparison can be done

    """
    template = '{2}{lightpurple}{0:{1}^50}{default}{2}'

    patch = PatchManager(oldtext, newtext, context=n, by_letter=by_letter)
    pywikibot.output(color_format(template, '  ALL CHANGES  ', '*', '\n'))

    for hunk in patch.hunks:
        pywikibot.output(hunk.diff_text)
    pywikibot.output(color_format(template, '  REVIEW CHANGES  ', '*', '\n'))

    text_list = patch.apply()
    pywikibot.output(color_format(template, '  APPROVED CHANGES  ', '*', '\n'))

    if any(hunk.reviewed == hunk.APPR for hunk in patch.hunks):
        for hunk in patch.hunks:
            if hunk.reviewed == hunk.APPR:
                pywikibot.output(hunk.diff_text)
    else:
        pywikibot.output(color_format(template, 'None.', '', ''))

    text = ''.join(text_list)

    return text


def html_comparator(compare_string):
    """List of added and deleted contexts from 'action=compare' html string.

    This function is useful when combineds with site.py's "compare" method.
    Site.compare() returns HTML that is useful for displaying on a page.
    Here we use BeautifulSoup to get the un-HTML-ify the context of changes.
    Finally we present the added and deleted contexts.
    @param compare_string: HTML string from mediawiki API
    @type compare_string: str
    @return: deleted and added list of contexts
    @rtype: dict
    """
    from bs4 import BeautifulSoup

    comparands = {'deleted-context': [], 'added-context': []}
    soup = BeautifulSoup(compare_string, 'html.parser')
    for change_type, css_class in (('deleted-context', 'diff-deletedline'),
                                   ('added-context', 'diff-addedline')):
        crutons = soup.find_all('td', class_=css_class)
        for cruton in crutons:
            cruton_string = ''.join(cruton.strings)
            comparands[change_type].append(cruton_string)
    return comparands



# -*- coding: utf-8 -*-
"""Family module for Omega Wiki."""
#
# (C) Pywikibot team, 2006-2019
#
# Distributed under the terms of the MIT license.
#
from __future__ import absolute_import, division, unicode_literals

from pywikibot import family


# Omegawiki, the Ultimate online dictionary
class Family(family.SingleSiteFamily):

    """Family class for Omega Wiki."""

    name = 'omegawiki'
    domain = 'www.omegawiki.org'

    # On most Wikipedias page names must start with a capital letter, but some
    # languages don't use this.
    nocapitalize = ['omegawiki']

    def scriptpath(self, code):
        """Return the script path for this family."""
        return ''
